@@ -32,7 +32,8 @@ from spyder.api.config.decorators import on_conf_change
 from spyder.api.translations import get_translation
 from spyder.api.widgets.main_widget import PluginMainWidget
 from spyder.api.widgets.menus import MENU_SEPARATOR
-from spyder.config.base import get_conf_path, running_under_pytest
+from spyder.config.base import (
+    get_conf_path, get_home_dir, running_under_pytest)
 from spyder.plugins.ipythonconsole.utils.kernelspec import SpyderKernelSpec
 from spyder.plugins.ipythonconsole.utils.manager import SpyderKernelManager
 from spyder.plugins.ipythonconsole.utils.ssh import openssh_tunnel
@@ -65,9 +66,9 @@ SPYDER_KERNELS_VERSION_MSG = _(
     '>= {0} and < {1}').format(
         SPYDER_KERNELS_MIN_VERSION, SPYDER_KERNELS_MAX_VERSION)
 SPYDER_KERNELS_CONDA = (
-    f'conda install spyder-kernels={SPYDER_KERNELS_MIN_VERSION[:-2]}')
+    f'conda install spyder&#45;kernels={SPYDER_KERNELS_MIN_VERSION[:-2]}')
 SPYDER_KERNELS_PIP = (
-    f'pip install spyder-kernels=={SPYDER_KERNELS_MIN_VERSION[:-1]}*')
+    f'pip install spyder&#45;kernels=={SPYDER_KERNELS_MIN_VERSION[:-1]}*')
 
 
 class IPythonConsoleWidgetActions:
@@ -392,11 +393,15 @@ class IPythonConsoleWidget(PluginMainWidget):
         # See spyder-ide/spyder#11880
         self._init_asyncio_patch()
 
+        # To cache kernel properties
         self._cached_kernel_properties = None
 
+        # Initial value for the current working directory
+        self._current_working_directory = get_home_dir()
+
     def on_close(self):
         self.mainwindow_close = True
-        self.close_clients()
+        self.close_all_clients()
 
     # ---- PluginMainWidget API and settings handling
     # ------------------------------------------------------------------------
@@ -442,7 +447,7 @@ class IPythonConsoleWidget(PluginMainWidget):
             text=_("Connect to an existing kernel"),
             tip=_("Open a new IPython console connected to an existing "
                   "kernel"),
-            triggered=self.create_client_for_kernel,
+            triggered=self._create_client_for_kernel,
         )
         self.rename_tab_action = self.create_action(
             IPythonConsoleWidgetActions.RenameTab,
@@ -662,8 +667,11 @@ class IPythonConsoleWidget(PluginMainWidget):
                 client.get_control().set_help_enabled,
                 value)
 
-    @on_conf_change(section='appearance', option='selected')
-    def change_clients_color_scheme(self, value):
+    @on_conf_change(section='appearance', option=['selected', 'ui_theme'])
+    def change_clients_color_scheme(self, option, value):
+        if option == 'ui_theme':
+            value = self.get_conf('selected', section='appearance')
+
         for idx, client in enumerate(self.clients):
             self._change_client_conf(
                 client,
@@ -1075,155 +1083,16 @@ class IPythonConsoleWidget(PluginMainWidget):
         except RuntimeError:
             pass
 
-    def _create_client_for_kernel(self, connection_file, hostname, sshkey,
-                                  password):
-        # Verifying if the connection file exists
-        try:
-            cf_path = osp.dirname(connection_file)
-            cf_filename = osp.basename(connection_file)
-            # To change a possible empty string to None
-            cf_path = cf_path if cf_path else None
-            connection_file = find_connection_file(filename=cf_filename,
-                                                   path=cf_path)
-            if os.path.splitext(connection_file)[1] != ".json":
-                # There might be a file with the same id in the path.
-                connection_file = find_connection_file(
-                    filename=cf_filename + ".json", path=cf_path)
-        except (IOError, UnboundLocalError):
-            QMessageBox.critical(self, _('IPython'),
-                                 _("Unable to connect to "
-                                   "<b>%s</b>") % connection_file)
+    @Slot()
+    def _create_client_for_kernel(self):
+        """Create a client connected to an existing kernel"""
+        connect_output = KernelConnectionDialog.get_connection_parameters(self)
+        (connection_file, hostname, sshkey, password, ok) = connect_output
+        if not ok:
             return
-
-        # Getting the master id that corresponds to the client
-        # (i.e. the i in i/A)
-        master_id = None
-        given_name = None
-        is_external_kernel = True
-        known_spyder_kernel = False
-        slave_ord = ord('A') - 1
-        kernel_manager = None
-        stderr_obj = None
-        stdout_obj = None
-        fault_obj = None
-
-        for cl in self.clients:
-            if connection_file in cl.connection_file:
-                if cl.get_kernel() is not None:
-                    kernel_manager = cl.get_kernel()
-                connection_file = cl.connection_file
-                if master_id is None:
-                    master_id = cl.id_['int_id']
-                    is_external_kernel = cl.shellwidget.is_external_kernel
-                    known_spyder_kernel = cl.shellwidget.is_spyder_kernel
-                    if cl.stderr_obj:
-                        stderr_obj = cl.stderr_obj.copy()
-                    if cl.stdout_obj:
-                        stdout_obj = cl.stdout_obj.copy()
-                    if cl.fault_obj:
-                        fault_obj = cl.fault_obj.copy()
-                given_name = cl.given_name
-                new_slave_ord = ord(cl.id_['str_id'])
-                if new_slave_ord > slave_ord:
-                    slave_ord = new_slave_ord
-
-        # If we couldn't find a client with the same connection file,
-        # it means this is a new master client
-        if master_id is None:
-            self.master_clients += 1
-            master_id = str(self.master_clients)
-
-        # Set full client name
-        client_id = dict(int_id=master_id,
-                         str_id=chr(slave_ord + 1))
-
-        # Creating the client
-        show_elapsed_time = self.get_conf('show_elapsed_time')
-        reset_warning = self.get_conf('show_reset_namespace_warning')
-        ask_before_restart = self.get_conf('ask_before_restart')
-        client = ClientWidget(self,
-                              id_=client_id,
-                              given_name=given_name,
-                              history_filename=get_conf_path('history.py'),
-                              config_options=self.config_options(),
-                              additional_options=self.additional_options(),
-                              interpreter_versions=self.interpreter_versions(),
-                              connection_file=connection_file,
-                              context_menu_actions=self.context_menu_actions,
-                              time_label=self.time_label,
-                              hostname=hostname,
-                              is_external_kernel=is_external_kernel,
-                              is_spyder_kernel=known_spyder_kernel,
-                              show_elapsed_time=show_elapsed_time,
-                              reset_warning=reset_warning,
-                              ask_before_restart=ask_before_restart,
-                              css_path=self.css_path,
-                              handlers=self.registered_spyder_kernel_handlers,
-                              stderr_obj=stderr_obj,
-                              stdout_obj=stdout_obj,
-                              fault_obj=fault_obj)
-
-        # Create kernel client
-        kernel_client = QtKernelClient(connection_file=connection_file)
-
-        # This is needed for issue spyder-ide/spyder#9304.
-        try:
-            kernel_client.load_connection_file()
-        except Exception as e:
-            QMessageBox.critical(self, _('Connection error'),
-                                 _("An error occurred while trying to load "
-                                   "the kernel connection file. The error "
-                                   "was:\n\n") + str(e))
-            return
-
-        if hostname is not None:
-            try:
-                connection_info = dict(
-                    ip=kernel_client.ip,
-                    shell_port=kernel_client.shell_port,
-                    iopub_port=kernel_client.iopub_port,
-                    stdin_port=kernel_client.stdin_port,
-                    hb_port=kernel_client.hb_port)
-                newports = self.tunnel_to_kernel(connection_info, hostname,
-                                                 sshkey, password)
-                (kernel_client.shell_port,
-                 kernel_client.iopub_port,
-                 kernel_client.stdin_port,
-                 kernel_client.hb_port) = newports
-                # Save parameters to connect comm later
-                kernel_client.ssh_parameters = (hostname, sshkey, password)
-            except Exception as e:
-                QMessageBox.critical(self, _('Connection error'),
-                                     _("Could not open ssh tunnel. The "
-                                       "error was:\n\n") + str(e))
-                return
-
-        # Assign kernel manager and client to shellwidget
-        kernel_client.start_channels()
-        shellwidget = client.shellwidget
-        shellwidget.set_kernel_client_and_manager(
-            kernel_client, kernel_manager)
-        shellwidget.sig_exception_occurred.connect(
-            self.sig_exception_occurred)
-
-        if not known_spyder_kernel:
-            shellwidget.sig_is_spykernel.connect(
-                self.connect_external_spyder_kernel)
-            shellwidget.check_spyder_kernel()
-
-        self.sig_shellwidget_created.emit(shellwidget)
-        kernel_client.stopped_channels.connect(
-            lambda: self.sig_shellwidget_deleted.emit(shellwidget))
-
-        # Set elapsed time, if possible
-        if not is_external_kernel:
-            self.set_client_elapsed_time(client)
-
-        # Adding a new tab for the client
-        self.add_tab(client, name=client.get_name())
-
-        # Register client
-        self.register_client(client)
+        else:
+            self.create_client_for_kernel(connection_file, hostname, sshkey,
+                                          password)
 
     # ---- Public API
     # -------------------------------------------------------------------------
@@ -1295,6 +1164,14 @@ class IPythonConsoleWidget(PluginMainWidget):
                 sw.is_waiting_pdb_input(), sw.get_pdb_last_step())
             self.sig_shellwidget_changed.emit(sw)
 
+            # This is necessary to sync the current client cwd with the working
+            # directory displayed by other plugins in Spyder (e.g. Files).
+            # NOTE: Instead of emitting sig_current_directory_changed directly,
+            # we call on_working_directory_changed to validate that the cwd
+            # exists (this couldn't be the case for remote kernels).
+            if sw.get_cwd() != self.get_working_directory():
+                self.on_working_directory_changed(sw.get_cwd())
+
         self.update_tabs_text()
         self.update_actions()
 
@@ -1346,7 +1223,7 @@ class IPythonConsoleWidget(PluginMainWidget):
             pass
 
     def rename_client_tab(self, client, given_name):
-        """Rename client's tab."""
+        """Rename a client's tab."""
         index = self.get_client_index_from_id(id(client))
 
         if given_name is not None:
@@ -1542,14 +1419,15 @@ class IPythonConsoleWidget(PluginMainWidget):
     @Slot(bool, bool)
     @Slot(bool, str, bool)
     def create_new_client(self, give_focus=True, filename='', is_cython=False,
-                          is_pylab=False, is_sympy=False, given_name=None):
+                          is_pylab=False, is_sympy=False, given_name=None,
+                          cache=True, initial_cwd=None):
         """Create a new client"""
         self.master_clients += 1
         client_id = dict(int_id=str(self.master_clients),
                          str_id='A')
         std_dir = self._test_dir if self._test_dir else None
         cf, km, kc, stderr_obj, stdout_obj = self.get_new_kernel(
-            is_cython, is_pylab, is_sympy, std_dir=std_dir)
+            is_cython, is_pylab, is_sympy, std_dir=std_dir, cache=cache)
 
         if cf is not None:
             fault_obj = StdFile(cf, '.fault', std_dir)
@@ -1569,7 +1447,6 @@ class IPythonConsoleWidget(PluginMainWidget):
                               interpreter_versions=self.interpreter_versions(),
                               connection_file=cf,
                               context_menu_actions=self.context_menu_actions,
-                              time_label=self.time_label,
                               show_elapsed_time=show_elapsed_time,
                               reset_warning=reset_warning,
                               given_name=given_name,
@@ -1580,7 +1457,8 @@ class IPythonConsoleWidget(PluginMainWidget):
                               handlers=self.registered_spyder_kernel_handlers,
                               stderr_obj=stderr_obj,
                               stdout_obj=stdout_obj,
-                              fault_obj=fault_obj)
+                              fault_obj=fault_obj,
+                              initial_cwd=initial_cwd)
 
         self.add_tab(
             client, name=client.get_name(), filename=filename,
@@ -1631,6 +1509,157 @@ class IPythonConsoleWidget(PluginMainWidget):
         if client.shellwidget.kernel_manager is None:
             return
         self.register_client(client, give_focus=give_focus)
+        return client
+
+    def create_client_for_kernel(self, connection_file, hostname, sshkey,
+                                 password):
+        """Create a client connected to an existing kernel."""
+        # Verifying if the connection file exists
+        try:
+            cf_path = osp.dirname(connection_file)
+            cf_filename = osp.basename(connection_file)
+            # To change a possible empty string to None
+            cf_path = cf_path if cf_path else None
+            connection_file = find_connection_file(filename=cf_filename,
+                                                   path=cf_path)
+            if os.path.splitext(connection_file)[1] != ".json":
+                # There might be a file with the same id in the path.
+                connection_file = find_connection_file(
+                    filename=cf_filename + ".json", path=cf_path)
+        except (IOError, UnboundLocalError):
+            QMessageBox.critical(self, _('IPython'),
+                                 _("Unable to connect to "
+                                   "<b>%s</b>") % connection_file)
+            return
+
+        # Getting the master id that corresponds to the client
+        # (i.e. the i in i/A)
+        master_id = None
+        given_name = None
+        is_external_kernel = True
+        known_spyder_kernel = False
+        slave_ord = ord('A') - 1
+        kernel_manager = None
+        stderr_obj = None
+        stdout_obj = None
+        fault_obj = None
+
+        for cl in self.clients:
+            if connection_file in cl.connection_file:
+                if cl.get_kernel() is not None:
+                    kernel_manager = cl.get_kernel()
+                connection_file = cl.connection_file
+                if master_id is None:
+                    master_id = cl.id_['int_id']
+                    is_external_kernel = cl.shellwidget.is_external_kernel
+                    known_spyder_kernel = cl.shellwidget.is_spyder_kernel
+                    if cl.stderr_obj:
+                        stderr_obj = cl.stderr_obj.copy()
+                    if cl.stdout_obj:
+                        stdout_obj = cl.stdout_obj.copy()
+                    if cl.fault_obj:
+                        fault_obj = cl.fault_obj.copy()
+                given_name = cl.given_name
+                new_slave_ord = ord(cl.id_['str_id'])
+                if new_slave_ord > slave_ord:
+                    slave_ord = new_slave_ord
+
+        # If we couldn't find a client with the same connection file,
+        # it means this is a new master client
+        if master_id is None:
+            self.master_clients += 1
+            master_id = str(self.master_clients)
+
+        # Set full client name
+        client_id = dict(int_id=master_id,
+                         str_id=chr(slave_ord + 1))
+
+        # Creating the client
+        show_elapsed_time = self.get_conf('show_elapsed_time')
+        reset_warning = self.get_conf('show_reset_namespace_warning')
+        ask_before_restart = self.get_conf('ask_before_restart')
+        client = ClientWidget(self,
+                              id_=client_id,
+                              given_name=given_name,
+                              history_filename=get_conf_path('history.py'),
+                              config_options=self.config_options(),
+                              additional_options=self.additional_options(),
+                              interpreter_versions=self.interpreter_versions(),
+                              connection_file=connection_file,
+                              context_menu_actions=self.context_menu_actions,
+                              hostname=hostname,
+                              is_external_kernel=is_external_kernel,
+                              is_spyder_kernel=known_spyder_kernel,
+                              show_elapsed_time=show_elapsed_time,
+                              reset_warning=reset_warning,
+                              ask_before_restart=ask_before_restart,
+                              css_path=self.css_path,
+                              handlers=self.registered_spyder_kernel_handlers,
+                              stderr_obj=stderr_obj,
+                              stdout_obj=stdout_obj,
+                              fault_obj=fault_obj)
+
+        # Create kernel client
+        kernel_client = QtKernelClient(connection_file=connection_file)
+
+        # This is needed for issue spyder-ide/spyder#9304.
+        try:
+            kernel_client.load_connection_file()
+        except Exception as e:
+            QMessageBox.critical(self, _('Connection error'),
+                                 _("An error occurred while trying to load "
+                                   "the kernel connection file. The error "
+                                   "was:\n\n") + str(e))
+            return
+
+        if hostname is not None:
+            try:
+                connection_info = dict(
+                    ip=kernel_client.ip,
+                    shell_port=kernel_client.shell_port,
+                    iopub_port=kernel_client.iopub_port,
+                    stdin_port=kernel_client.stdin_port,
+                    hb_port=kernel_client.hb_port)
+                newports = self.tunnel_to_kernel(connection_info, hostname,
+                                                 sshkey, password)
+                (kernel_client.shell_port,
+                 kernel_client.iopub_port,
+                 kernel_client.stdin_port,
+                 kernel_client.hb_port) = newports
+                # Save parameters to connect comm later
+                kernel_client.ssh_parameters = (hostname, sshkey, password)
+            except Exception as e:
+                QMessageBox.critical(self, _('Connection error'),
+                                     _("Could not open ssh tunnel. The "
+                                       "error was:\n\n") + str(e))
+                return
+
+        # Assign kernel manager and client to shellwidget
+        kernel_client.start_channels()
+        shellwidget = client.shellwidget
+        shellwidget.set_kernel_client_and_manager(
+            kernel_client, kernel_manager)
+        shellwidget.sig_exception_occurred.connect(
+            self.sig_exception_occurred)
+
+        if not known_spyder_kernel:
+            shellwidget.sig_is_spykernel.connect(
+                self.connect_external_spyder_kernel)
+            shellwidget.check_spyder_kernel()
+
+        self.sig_shellwidget_created.emit(shellwidget)
+        kernel_client.stopped_channels.connect(
+            lambda: self.sig_shellwidget_deleted.emit(shellwidget))
+
+        # Set elapsed time, if possible
+        if not is_external_kernel:
+            self.set_client_elapsed_time(client)
+
+        # Adding a new tab for the client
+        self.add_tab(client, name=client.get_name())
+
+        # Register client
+        self.register_client(client)
 
     def create_pylab_client(self):
         """Force creation of Pylab client"""
@@ -1644,19 +1673,8 @@ class IPythonConsoleWidget(PluginMainWidget):
         """Force creation of Cython client"""
         self.create_new_client(is_cython=True, given_name="Cython")
 
-    @Slot()
-    def create_client_for_kernel(self):
-        """Create a client connected to an existing kernel"""
-        connect_output = KernelConnectionDialog.get_connection_parameters(self)
-        (connection_file, hostname, sshkey, password, ok) = connect_output
-        if not ok:
-            return
-        else:
-            self._create_client_for_kernel(connection_file, hostname, sshkey,
-                                           password)
-
     def get_new_kernel(self, is_cython=False, is_pylab=False,
-                       is_sympy=False, std_dir=None):
+                       is_sympy=False, std_dir=None, cache=True):
         """Get a new kernel, and cache one for next time."""
         # Cache another kernel for next time.
         kernel_spec = self.create_kernel_spec(
@@ -1666,8 +1684,9 @@ class IPythonConsoleWidget(PluginMainWidget):
         )
 
         new_kernel = self.create_new_kernel(kernel_spec, std_dir)
-        if new_kernel[2] is None:
-            # error
+
+        if new_kernel[2] is None or not cache:
+            # error or remove/don't use cache if requested
             self.close_cached_kernel()
             return new_kernel
 
@@ -1782,28 +1801,27 @@ class IPythonConsoleWidget(PluginMainWidget):
         # type is Qt.DirectConnection.
         self._shellwidget_started(client)
 
-
     @Slot(str)
     def create_client_from_path(self, path):
         """Create a client with its cwd pointing to path."""
-        self.create_new_client()
-        sw = self.get_current_shellwidget()
-        sw.set_cwd(path)
+        self.create_new_client(initial_cwd=path)
 
     def create_client_for_file(self, filename, is_cython=False):
         """Create a client to execute code related to a file."""
         # Create client
-        self.create_new_client(filename=filename, is_cython=is_cython)
+        client = self.create_new_client(
+            filename=filename, is_cython=is_cython)
 
         # Don't increase the count of master clients
         self.master_clients -= 1
 
         # Rename client tab with filename
-        client = self.get_current_client()
         client.allow_rename = False
         tab_text = self.disambiguate_fname(filename)
         self.rename_client_tab(client, tab_text)
 
+        return client
+
     def get_client_for_file(self, filename):
         """Get client associated with a given file."""
         client = None
@@ -1852,11 +1870,14 @@ class IPythonConsoleWidget(PluginMainWidget):
 
         # Connect to working directory
         shellwidget.sig_working_directory_changed.connect(
-            self.set_working_directory)
+            self.on_working_directory_changed)
 
         # Connect client execution state to be reflected in the interface
         client.sig_execution_state_changed.connect(self.update_actions)
 
+        # Show time label
+        client.sig_time_label.connect(self.time_label.setText)
+
     def close_client(self, index=None, client=None, ask_recursive=True):
         """Close client tab from index or widget (or close current tab)"""
         if not self.tabwidget.count():
@@ -1875,19 +1896,6 @@ class IPythonConsoleWidget(PluginMainWidget):
         if index is not None:
             client = self.tabwidget.widget(index)
 
-        # Needed to handle a RuntimeError. See spyder-ide/spyder#5568.
-        try:
-            # Close client
-            client.stop_button_click_handler()
-        except RuntimeError:
-            pass
-
-        # Disconnect timer needed to update elapsed time
-        try:
-            client.timer.timeout.disconnect(client.show_time)
-        except (RuntimeError, TypeError):
-            pass
-
         # Check if related clients or kernels are opened
         # and eventually ask before closing them
         if not self.mainwindow_close and ask_recursive:
@@ -1911,13 +1919,11 @@ class IPythonConsoleWidget(PluginMainWidget):
             if close_all == QMessageBox.Yes:
                 self.close_related_clients(client)
 
-        last_client = len(self.get_related_clients(client)) == 0
-        client.shutdown(last_client)
-        client.close()
-
         # Note: client index may have changed after closing related widgets
         self.tabwidget.removeTab(self.tabwidget.indexOf(client))
         self.clients.remove(client)
+        is_last_client = len(self.get_related_clients(client)) == 0
+        client.close_client(is_last_client)
 
         # This is needed to prevent that hanged consoles make reference
         # to an index that doesn't exist. See spyder-ide/spyder#4881
@@ -1932,7 +1938,7 @@ class IPythonConsoleWidget(PluginMainWidget):
         if not self.tabwidget.count() and self.create_new_client_if_empty:
             self.create_new_client()
 
-    def close_clients(self):
+    def close_all_clients(self):
         """
         Perform close actions for each running client.
 
@@ -1941,17 +1947,23 @@ class IPythonConsoleWidget(PluginMainWidget):
         bool
             If the closing action was succesful.
         """
+        # IMPORTANT: **Do not** change this way of closing clients, which uses
+        # a copy of `self.clients`, because it preserves the main window layout
+        # when Spyder is closed.
+        # Fixes spyder-ide/spyder#19084
         open_clients = self.clients.copy()
         for client in self.clients:
-            last_client = (
+            is_last_client = (
                 len(self.get_related_clients(client, open_clients)) == 0)
-            client.shutdown(last_client)
-            client.close()
+            client.close_client(is_last_client)
             open_clients.remove(client)
+
         # Close all closing shellwidgets.
         ShellWidget.wait_all_shutdown()
+
         # Close cached kernel
         self.close_cached_kernel()
+        self.filenames = []
         return True
 
     def get_client_index_from_id(self, client_id):
@@ -1991,7 +2003,7 @@ class IPythonConsoleWidget(PluginMainWidget):
         for i in range(len(self.clients)):
             client = self.clients[-1]
             self.close_client(client=client, ask_recursive=False)
-        self.create_new_client(give_focus=False)
+        self.create_new_client(give_focus=False, cache=False)
         self.create_new_client_if_empty = True
 
     def current_client_inspect_object(self):
@@ -2169,25 +2181,30 @@ class IPythonConsoleWidget(PluginMainWidget):
         # editor during repeated, rapid entry of debugging commands.
         self.sig_edit_goto_requested[str, int, str, bool].emit(
             fname, lineno, '', False)
-        self.activateWindow()
-        shellwidget._control.setFocus()
+
+        # Give focus to console if requested
+        if shellwidget._pdb_focus_to_editor:
+            # Next focus to editor was enabled
+            shellwidget._pdb_focus_to_editor = False
+        else:
+            self.activateWindow()
+            shellwidget._control.setFocus()
 
     def set_spyder_breakpoints(self):
         """Set Spyder breakpoints into all clients"""
         for cl in self.clients:
             cl.shellwidget.set_spyder_breakpoints()
 
-    def pdb_execute_command(self, command):
+    def pdb_execute_command(self, command, focus_to_editor):
         """
         Send command to the pdb kernel if possible.
         """
-
         sw = self.get_current_shellwidget()
         if sw is not None:
             # Needed to handle an error when kernel_client is None.
             # See spyder-ide/spyder#7578.
             try:
-                sw.pdb_execute_command(command)
+                sw.pdb_execute_command(command, focus_to_editor)
             except AttributeError:
                 pass
 
@@ -2255,6 +2272,9 @@ class IPythonConsoleWidget(PluginMainWidget):
                                 repr(cell_name),
                                 norm(filename).replace("'", r"\'")))
 
+                if function == "debugcell":
+                    # To keep focus in editor after running debugfile
+                    client.shellwidget._pdb_focus_to_editor = focus_to_editor
             # External kernels and run_cell_copy, just execute the code
             else:
                 line = code.strip()
@@ -2266,7 +2286,7 @@ class IPythonConsoleWidget(PluginMainWidget):
 
             # This is necessary to prevent raising the console if the editor
             # and console are tabified next to each other and the 'Maintain
-            # focus in the editor' option is activated.
+            # focus in the editor' option is enabled.
             # Fixes spyder-ide/spyder#17028
             if not focus_to_editor:
                 self.sig_switch_to_plugin_requested.emit()
@@ -2289,7 +2309,8 @@ class IPythonConsoleWidget(PluginMainWidget):
 
     # ---- For scripts
     def run_script(self, filename, wdir, args, debug, post_mortem,
-                   current_client, clear_variables, console_namespace):
+                   current_client, clear_variables, console_namespace,
+                   focus_to_editor):
         """Run script in current or dedicated client"""
         norm = lambda text: remove_backslashes(str(text))
 
@@ -2305,8 +2326,8 @@ class IPythonConsoleWidget(PluginMainWidget):
         else:
             client = self.get_client_for_file(filename)
             if client is None:
-                self.create_client_for_file(filename, is_cython=is_cython)
-                client = self.get_current_client()
+                client = self.create_client_for_file(
+                    filename, is_cython=is_cython)
                 is_new_client = True
 
         if client is not None:
@@ -2323,6 +2344,10 @@ class IPythonConsoleWidget(PluginMainWidget):
                 if console_namespace:
                     line += ", current_namespace=True"
                 line += ")"
+
+                if debug:
+                    # To keep focus in editor after running debugfile
+                    client.shellwidget._pdb_focus_to_editor = focus_to_editor
             else:  # External, non spyder-kernels, use %run
                 line = "%run "
                 if debug:
@@ -2339,19 +2364,25 @@ class IPythonConsoleWidget(PluginMainWidget):
                     pass
                 elif current_client:
                     self.execute_code(line, current_client, clear_variables,
-                                      set_focus=False)
+                                      set_focus=not focus_to_editor)
                 else:
                     if is_new_client:
                         client.shellwidget.silent_execute('%clear')
                     else:
                         client.shellwidget.execute('%clear')
                     client.shellwidget.sig_prompt_ready.connect(
-                            lambda: self.execute_code(
-                                line, current_client, clear_variables,
-                                set_focus=False))
+                        lambda: self.execute_code(
+                            line, current_client, clear_variables,
+                            set_focus=not focus_to_editor,
+                            shellwidget=client.shellwidget
+                        )
+                    )
             except AttributeError:
                 pass
-            self.sig_switch_to_plugin_requested.emit()
+
+            # Show plugin after execution if focus is going to be given to it.
+            if not focus_to_editor:
+                self.sig_switch_to_plugin_requested.emit()
         else:
             # XXX: not sure it can really happen
             QMessageBox.warning(
@@ -2364,24 +2395,39 @@ class IPythonConsoleWidget(PluginMainWidget):
             )
 
     # ---- For working directory and path management
+    def set_working_directory(self, dirname):
+        """
+        Set current working directory in the Working Directory and Files
+        plugins.
+        """
+        if osp.isdir(dirname):
+            self.sig_current_directory_changed.emit(dirname)
+
+    def save_working_directory(self, dirname):
+        """
+        Save current working directory when changed by the Working Directory
+        plugin.
+        """
+        self._current_working_directory = dirname
+
+    def get_working_directory(self):
+        """Get saved value of current working directory."""
+        return self._current_working_directory
+
     def set_current_client_working_directory(self, directory):
         """Set current client working directory."""
         shellwidget = self.get_current_shellwidget()
         if shellwidget is not None:
             shellwidget.set_cwd(directory)
 
-    def set_working_directory(self, dirname):
+    def on_working_directory_changed(self, dirname):
         """
-        Set current working directory in the workingdirectory and explorer
-        plugins.
+        Notify that the working directory was changed in the current console
+        to other plugins.
         """
-        if osp.isdir(dirname):
+        if dirname and osp.isdir(dirname):
             self.sig_current_directory_changed.emit(dirname)
 
-    def get_working_directory(self):
-        """Get current working directory."""
-        return self.get_plugin()._get_working_directory()
-
     def update_working_directory(self):
         """Update working directory to console cwd."""
         shellwidget = self.get_current_shellwidget()
@@ -2417,9 +2463,12 @@ class IPythonConsoleWidget(PluginMainWidget):
 
     # ---- For execution
     def execute_code(self, lines, current_client=True, clear_variables=False,
-                     set_focus=True):
+                     set_focus=True, shellwidget=None):
         """Execute code instructions."""
-        sw = self.get_current_shellwidget()
+        if current_client:
+            sw = self.get_current_shellwidget()
+        else:
+            sw = shellwidget
         if sw is not None:
             if not current_client:
                 # Clear console and reset namespace for
@@ -2432,15 +2481,26 @@ class IPythonConsoleWidget(PluginMainWidget):
                 sw.reset_namespace(warning=False)
             elif current_client and clear_variables:
                 sw.reset_namespace(warning=False)
+
             # Needed to handle an error when kernel_client is none.
             # See spyder-ide/spyder#6308.
             try:
                 sw.execute(str(lines))
             except AttributeError:
                 pass
-            self.activateWindow()
+
             if set_focus:
-                self.get_current_client().get_control().setFocus()
+                # The `activateWindow` call below needs to be inside this `if`
+                # to avoid giving focus to the console when it's undocked,
+                # users are running code from the editor and the 'Maintain
+                # focus in the editor' option is enabled.
+                # Fixes spyder-ide/spyder#3221
+                self.activateWindow()
+
+                # Gives focus to the current client
+                focus_widget = self.get_focus_widget()
+                if focus_widget:
+                    focus_widget.setFocus()
 
     # ---- For error handling
     def go_to_error(self, text):
